/*
Copyright 2018-2019 The Doctl Authors All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package doctl

import "fmt"

// MissingArgsErr is returned when there are too few arguments for a command.
type MissingArgsErr struct {
	Command string
}

var _ error = &MissingArgsErr{}

// NewMissingArgsErr creates a MissingArgsErr instance.
func NewMissingArgsErr(cmd string) *MissingArgsErr {
	return &MissingArgsErr{Command: cmd}
}

func (e *MissingArgsErr) Error() string {
	return fmt.Sprintf("(%s) command is missing required arguments", e.Command)
}

// TooManyArgsErr is returned when there are too many arguments for a command.
type TooManyArgsErr struct {
	Command string
}

var _ error = &TooManyArgsErr{}

// NewTooManyArgsErr creates a TooManyArgsErr instance.
func NewTooManyArgsErr(cmd string) *TooManyArgsErr {
	return &TooManyArgsErr{Command: cmd}
}

func (e *TooManyArgsErr) Error() string {
	return fmt.Sprintf("(%s) command contains unsupported arguments", e.Command)
}
